/**
  OpenGL/GLUT "playground" for 2D terrain generation.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-04-23 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include <vector>
#include "osl/vec4.h" /* 3D vector class */

#include "ogl/minicam.h" /* ortho_frame, camera motion, key_down & events */

/* Random Float: Return a random number between 0 and 1 */
float rf(void) { return (rand()&0xfff)*(1.0/0xfff); }
/* Random Vector: return a random vector in 3D, all axes between 0 and 1 */
vec3 rv(void) {return vec3(rf(),rf(),rf());}

/* These variables are changed directly by the user interface code below */

/* Enable for anisotropic filtering */
int texaniso=1;

/* Texture pixel format on graphics card */
int texformat=GL_RGBA8; 

/* Texture "wrap mode" in X and Y: */
int wrap_mode=GL_REPEAT;

/* Texture magnification mode/minification mode */
int minfilter_mode=GL_LINEAR_MIPMAP_LINEAR; 
int magfilter_mode=GL_LINEAR; 

/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %d at %s\n",err,where);
		exit(1);
	}
}

const unsigned int plasma_size=256;
const unsigned int plasma_mask=plasma_size-1;

/* Tiny little 2D "height image" class */
class terrain {
public:
	/* Size of our height image, in pixels. */
	int w,h;
	
	/* Data in our image: w * h pixels, each one a height. */
	float *ht;
	
	terrain(int w_,int h_) :w(w_), h(h_), ht(new float[w*h]) {}
	~terrain() {delete[] ht;}
	
	/* Access the terrain height at pixel (x,y) */
	float &operator() (int x,int y) { 
		x=x&plasma_mask; y=y&plasma_mask; /* wraparound */
		return ht[y*w+x];
	}
	
	/* Interpolate the height at this float location */
	float operator() (float x,float y) { 
		int ix=(int)x, iy=(int)y; /* integer corners */
		float dx=x-ix, dy=y-iy; /* location inside our pixel */
		float TL=(*this)(ix,iy+1), TR=(*this)(ix+1,iy+1);
		float BL=(*this)(ix,iy),   BR=(*this)(ix+1,iy);
		float T=TL+dx*(TR-TL); /* horizontal interpolations */
		float B=BL+dx*(BR-BL);
		return B+dy*(T-B); /* vertical interpolations */
	}
	
};

/**
 Random number generator: at the location X,Y (MUST be deterministic!),
 generate a perturbation for a span of "size" pixels.
*/
float perturb(unsigned int X,unsigned int Y, float size) {
	X&=plasma_mask; Y&=plasma_mask; /* makes image wrap properly */
	unsigned int seed=X*2444u + Y*98023u;
	unsigned int seed2=X*24943u + Y*980237u;
	unsigned int val=(seed^(seed2>>8)^(seed>>16)) % 256u;
	float scale=1.0/128; /* fixes magnitude of result */
	
	float falloff=1.0; /* decreases perturbation at small sizes */
	int falloffsize=100;
	if (size<falloffsize) falloff=size*(1.0/falloffsize);

	return 0.3*falloff*scale*(val-128.0);
}

/** Initialize terrain to v */
void init(terrain &t,float v) {
	for (int y=0;y<t.h;y++)
	for (int x=0;x<t.w;x++)
	{
		t(x,y)=v;
	}
}
/* Copy src terrain over dest */
void copy(terrain &dest,terrain &src) {
	for (int y=0;y<dest.h;y++)
	for (int x=0;x<dest.w;x++)
	{
		dest(x,y)=src(x,y);
	}
}
/** Blur the pixels in this terrain.  Uses red-black pixel pattern */
void blur(terrain &t) {
	for (int y=0;y<t.h;y++)
	for (int x=y&1;x<t.w;x+=2)
	{
		t(x,y)=0.25*(t(x-1,y)+t(x+1,y)+t(x,y+1)+t(x,y-1));
	}
	for (int y=0;y<t.h;y++)
	for (int x=!(y&1);x<t.w;x+=2)
	{
		t(x,y)=0.25*(t(x-1,y)+t(x+1,y)+t(x,y+1)+t(x,y-1));
	}
}

/** Recursively draw plasma fractal on this terrain. 
The four corners must already have been set.
       T
   L       R
       B
*/
void plasma(terrain &t,int L,int R, int T, int B) {
	if (R-L<=1) return;
	if (B-T<=1) return; /* no pixels in between to render! */
	
	// Read terrain corner values
	float TL=t(L,T);
	float TR=t(R,T);
	float BL=t(L,B);
	float BR=t(R,B);
	
	// Compute center coordinates
	int X=(L+R)/2, Y=(T+B)/2;
	
	// Compute perturbed centers
	float YL=(TL+BL)*0.5 + perturb(Y,L,B-T);
	float YR=(TR+BR)*0.5 + perturb(Y,R,B-T);
	float TX=(TL+TR)*0.5 + perturb(T,X,R-L);
	float BX=(BL+BR)*0.5 + perturb(B,X,R-L);
	float YX=(TL+TR+BL+BR)*0.25 + perturb(Y,X,(B-T + R-L)*(1.0/1.41));
	
	// Write back perturbed values onto terrain
	t(L,Y) = YL;
	t(R,Y) = YR;
	t(X,T) = TX;
	t(X,B) = BX;
	t(X,Y) = YX;
	
	// Recurse over four corners
	plasma(t,L,X, T,Y);
	plasma(t,L,X, Y,B);
	plasma(t,X,R, T,Y);
	plasma(t,X,R, Y,B);
}


class particle {
public:
	vec3 P; /* position */
	vec3 V; /* velocity */
	float dirt; /* suspended terrain material */
};


void display(void) /* Draw everything in our window */
{
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float display_dt=0.0;
	float big=3; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	static terrain t(plasma_size+1,plasma_size+1); /* rock */
	static terrain water(plasma_size+1,plasma_size+1); /* water */
	static GLuint srcTex=0;
	if (srcTex==0) 
	{ /* Set up the terrain */
		init(t,0.5); /* init to 0.5 */
		init(water,0.0); /* init to 0.0 */
		plasma(t,0,t.w-1, 0,t.h-1);
		
		glGenTextures(1,&srcTex);
		glBindTexture(GL_TEXTURE_2D,srcTex);
		glTexImage2D(GL_TEXTURE_2D,0,
			GL_RGBA8, t.w,t.h,
			0, GL_LUMINANCE,GL_FLOAT,&t(0,0));
		glTexParameteri(GL_TEXTURE_2D,
			GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	}
  
/* Main camera setup */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.01,1000.0 /* Z range */
	);
	
	/* Physics and display timesteps: */
	oglCameraLookAt(0.5);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
// Draw in our ground plane texture
	glBindTexture(GL_TEXTURE_2D,srcTex);
	glGenerateMipmapEXT(GL_TEXTURE_2D); /* make mipmaps before rendering */
	glEnable(GL_TEXTURE_2D);
	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,magfilter_mode); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,minfilter_mode);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap_mode); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap_mode);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT,texaniso?128:1);

	glDisable(GL_BLEND);
	glBegin(GL_QUADS);
	glColor4f(1,1,1,1); 
	glTexCoord2f(-big,-big); glVertex3f(-big,-big,floor);
	glTexCoord2f(+big,-big); glVertex3f(+big,-big,floor);
	glTexCoord2f(+big,+big); glVertex3f(+big,+big,floor);
	glTexCoord2f(-big,+big); glVertex3f(-big,+big,floor);
	glEnd();
	
	glEnable(GL_DEPTH_TEST);
	
	/* Draw the terrain in 3D */
	glBegin(GL_TRIANGLE_STRIP);
	double xs=1.0/t.w, ys=1.0/t.h;
	float ts=0.5; /* scale from t values to heights */
	double terraindelta=0.001; /* Z shift (avoid Z fighting w/ water) */
	for (int y=0;y<t.h-1;y++) {
		vec3 p;
		for (int x=0;x<t.w;x++) {
			p=vec3(x*xs,y*ys,(t(x,y))*ts-terraindelta); /* top */
			glTexCoord2fv(p); glVertex3fv(p);
			if (x==0) glVertex3fv(p); /* duplicate first vertex! */
			
			p=vec3(x*xs,(y+1)*ys,(t(x,y+1))*ts-terraindelta); /* bottom */
			glTexCoord2fv(p); glVertex3fv(p);
		}
		glVertex3fv(p); /* duplicate vertex (to split triangle strip) */
	}
	glEnd();
	
	glDisable(GL_TEXTURE_2D);
	

/* Simulate particles of water */
	static std::vector<particle> pts;
	if (pts.size()==0) { /* First time: initialize points */
		srand(1);
		for (int i=0;i<10000;i++) {
			particle p;
			p.P=0.1*vec3(rf(),rf(),0.0)+vec3(0.5,0.5,0);
			p.V=0.00003*(rv()-vec3(0.5)); 
			p.dirt=0.0;
			pts.push_back(p);
		}
	}
	
	/* Subcycle, so the physics runs at the right stepsize, 
	   yet still keeps up with the display. */
	static double sim_time=0.0, dt=0.01;
	double wall_time=oglTimer()*1.0; /* <- run physics at multiple of realtime */
	while (sim_time<wall_time && !key_down[(int)' '])
	{ /* take one physics step */
		sim_time+=dt;
		
		/* Recycle a few particles to the top of the mountain */
		for (int cyc=0;cyc<4;cyc++) {
			particle &p=pts[rand()%pts.size()];
			p.P=vec3(0.38,0.75,0.0);
			p.dirt=0.0; /* don't teleport in dirt... */
		}
		
		/* Figure out where there's water */
		copy(water,t); /* water slides off terrain */
		for (int i=0;i<pts.size();i++) {
			particle &p=pts[i];
			int x=p.P.x*t.w, y=p.P.y*t.h;
			water(x,y)+=0.02; /* water repels other water */
		}
		/* blur the water field (on GPU, use mipmaps here!) */
		for (int b=0;b<1;b++) blur(water);
		
		/* Use the water field to apply particle forces */
		for (int i=0;i<pts.size();i++) {
			particle &p=pts[i];
			
			int x=p.P.x*t.w, y=p.P.y*t.h;
			int N=1; /* neighborhood to take gradient on */
			float C=water(x,y);
			float L=water(x-N,y);
			float R=water(x+N,y);
			float T=water(x,y+N);
			float B=water(x,y-N);
			vec3 grad(L-R,B-T,-1.0);
			float gs=1000.0; /* scale factor for gradient force */
			
		/* Apply physics */
			vec3 fA=-100.0*p.V; /* frictional acceleration (a=F/m) */
			vec3 A=grad*gs +fA; /* total acceleration */
			
			p.P+=dt*p.V+0.5*dt*dt*A; /* move position by velocity */
			p.V+=dt*A; /* accelerate particles downwards */
			
			float th=C*ts;
			if (p.P.z<th) {p.V.z=0.0; p.P.z=th; /* don't go below terrain */ }
			
			
		#if 1 /* Erosion! (doesn't seem to do anything very cool...) */
			float sol=0.02; /* amount of dirt that dissolves per timestep */
			float peaky=t(x,y)-0.25*(t(x-N,y)+t(x+N,y)+t(x,y-N)+t(x,y+N));
			float thresh=0.01;
			if (peaky>thresh) 
			{ /* we're on a local peak */
				if (p.dirt<1.0) { /* dissolve rock */
					t(x,y)-=sol;
					p.dirt=1.0;
				}
			}
			else if (peaky<-thresh) { /* we're in a local valley */
				if (p.dirt>0.0) { /* deposit rock */
					t(x,y)+=sol;
					p.dirt=0.0;
				}
			}
		#endif
		}
	}
	
	/* Draw the particles */
	glPointSize(4.0);
	glBegin(GL_POINTS);
	for (int i=0;i<pts.size();i++) {
		glColor3f(1,0,1);
		glVertex3fv(pts[i].P);
	}
	glEnd();
	
	
	
	oglCameraAxes();
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed+0.001;
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int main(int argc,char **argv) {
	glutInit(&argc,argv);  /* set up GLUT */
	
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	int win=glutCreateWindow("Fractal Terrain with 'Water'"); /* make a window */
	glewInit(); /* only AFTER create window! */
	
	glutDisplayFunc(display); /* here's how to draw the window */
	oglCameraInit();
	camera=vec3(0.5,0.5,0.80);
	camera_orient.y=vec3(0,1,0);camera_orient.z=vec3(0,0,1);
	glutMainLoop(); /* start drawing */
}
